/* -*- mode: c++ -*- */
#pragma once
#include <stdio.h>
#include <assert.h>
#include "common.h"

template <typename T>
class Worklist {
public:
  unsigned int sz;
  unsigned int *ndx;
  T *wl;

  Worklist(unsigned int size) {
    int zero = 0;

    check_cuda(cudaMalloc(&wl, size * sizeof(T)));

    sz = size;

    check_cuda(cudaMalloc(&ndx, 1 * sizeof(unsigned int)));
    check_cuda(cudaMemcpy(ndx, &zero, sizeof(unsigned int) * 1, cudaMemcpyHostToDevice));
    
  }

  void free() {
    check_cuda(cudaFree(wl));
  }

  __device__ void push(T item) {
    unsigned int pos;
   
    pos = *ndx + 1;

    // out of worklist space
    assert(pos < sz);

    wl[pos] = item;
  }

  void initialize(T *items, unsigned int n) {    
    check_cuda(cudaMemcpy(wl, items, sizeof(T) * n, cudaMemcpyHostToDevice));
    check_cuda(cudaMemcpy(ndx, &n, sizeof(*ndx) * 1, cudaMemcpyHostToDevice));  
  }

  unsigned int nitems() {
    unsigned int i;

    check_cuda(cudaMemcpy(&i, ndx, sizeof(unsigned int) * 1, cudaMemcpyDeviceToHost));

    return i;
  }

  void reset() {
    int zero = 0;

    check_cuda(cudaMemcpy(ndx, &zero, sizeof(unsigned int) * 1, cudaMemcpyHostToDevice));
  }

  __device__ int pop_index(unsigned int index, T &item) {
    if(index < *ndx) {
      item = wl[index];
      return 1;
    }

    return 0;
  }
};
